#version 430 core

in VertexData {
	vec3 position_world;
	vec3 normal_world;
	vec2 uv; //TexCoords
} vert;

out vec4 FragColor;

uniform vec3 camera_world; //camPos

uniform sampler2D texture_albedo;

//uniform vec3 albedo;
uniform float metallic;
uniform float roughness;
uniform float ao;
uniform float brightness;

const float PI = 3.14159265359;

uniform vec3 lightPositions[4];
uniform vec3 lightColors[4];

uniform samplerCube depthMap;
uniform float far_plane;
uniform int pattern = 0;

// array of offset direction for sampling
vec3 gridSamplingDisk[20] = vec3[]
(
   vec3(1, 1,  1), vec3( 1, -1,  1), vec3(-1, -1,  1), vec3(-1, 1,  1), 
   vec3(1, 1, -1), vec3( 1, -1, -1), vec3(-1, -1, -1), vec3(-1, 1, -1),
   vec3(1, 1,  0), vec3( 1, -1,  0), vec3(-1, -1,  0), vec3(-1, 1,  0),
   vec3(1, 0,  1), vec3(-1,  0,  1), vec3( 1,  0, -1), vec3(-1, 0, -1),
   vec3(0, 1,  1), vec3( 0, -1,  1), vec3( 0, -1, -1), vec3( 0, 1, -1)
);



vec3 CheckerboardPattern(float size, vec3 color1, vec3 color2)
{
    ivec2 cell = ivec2(floor(vert.uv * size));
    bool isEven = (mod(cell.x + cell.y, 2) == 0);
    vec3 color = isEven ? color1 : color2;
    
    return color;
}

vec3 CoolPattern(float size, vec3 color1, vec3 color2)
{
    float angle = atan(vert.uv.y, vert.uv.x);
    float radius = length(vert.uv);
    
    float repetitions = size / (2.0 * PI * radius);
    
    bool isEven = (mod(int(repetitions * angle), 2) == 0);
    vec3 color = isEven ? color1 : color2;
    
    return color;
}

float hash(float x)
{
    return fract(sin(x) * 43758.5453);
}

float hash(vec2 p)
{
    return fract(sin(dot(p, vec2(37.0, 17.0))) * 43758.5453);
}

float perlinNoise(vec2 p)
{
    vec2 i = floor(p);
    vec2 f = fract(p);

    // Smooth interpolation
    vec2 u = f * f * (3.0 - 2.0 * f);

    // Permutation
    float a = hash(i);
    float b = hash(i + vec2(1.0, 0.0));
    float c = hash(i + vec2(0.0, 1.0));
    float d = hash(i + vec2(1.0, 1.0));

    // Gradients
    vec2 g0 = vec2(a, c);
    vec2 g1 = vec2(b, c);
    vec2 g2 = vec2(a, d);
    vec2 g3 = vec2(b, d);

    // Dot products
    float n0 = dot(g0, f - vec2(0.0, 0.0));
    float n1 = dot(g1, f - vec2(1.0, 0.0));
    float n2 = dot(g2, f - vec2(0.0, 1.0));
    float n3 = dot(g3, f - vec2(1.0, 1.0));

    // Interpolation
    vec2 smooth_i = u * u * (3.0 - 2.0 * u);
    return mix(mix(n0, n1, smooth_i.x), mix(n2, n3, smooth_i.x), smooth_i.y);
}



vec3 MarblePattern(float size, float scale, float turbulence, vec3 color1, vec3 color2)
{
    
    vec2 marbleUV = vert.uv * size;
    float noise = perlinNoise(vec2(marbleUV.x * scale, marbleUV.y * scale * turbulence));
    ivec2 cell = ivec2(floor(marbleUV));
  
    bool isEven = (mod(cell.x + cell.y, 2) == 0);
    vec3 color = isEven ? color1 : color2;
    color += noise * 0.1; // turbulence effect +-?
    
    return color;
}


//pure noise
vec3 OrganicPattern(vec2 uv, sampler2D textureSampler)
{
    
    vec2 scaledUV = uv * 30.0;
    vec2 offsetUV = uv * 10.0;
    vec2 offset1 = vec2(0.5, 0.2);
    vec2 offset2 = vec2(-0.3, 0.7);

    float noiseValue1 = perlinNoise(scaledUV + offset1);
    float noiseValue2 = perlinNoise(scaledUV + offset2);
    float randomValue1 = fract(sin(dot(uv, vec2(12.9898, 78.233))) * 43758.5453);
    float randomValue2 = fract(sin(dot(uv, vec2(39.3464, 29.731))) * 24634.6345);
    float combinedValue = noiseValue1 * 0.5 + noiseValue2 * 0.5;
    combinedValue = mix(combinedValue, randomValue1, 0.7);
    combinedValue = mix(combinedValue, randomValue2, 0.3);

    vec3 textureColor = texture2D(textureSampler, uv).rgb;
    vec3 color = textureColor * combinedValue;
    return color;
}


float WoodPattern(vec2 uv, float frequency, float amplitude, float rings, float graininess)
{
   
    float ringsValue = cos(uv.y * frequency + sin(uv.x * frequency) * amplitude);
    float noiseValue = perlinNoise(uv * graininess) * 2.0 - 1.0;
    float combinedValue = (ringsValue + noiseValue) / 2.0;
    float woodValue = (combinedValue + 1.0) / 2.0;
    float woodPattern = fract(woodValue * rings);
    return woodPattern;
}

vec3 WoodColorPattern(vec2 uv, float frequency, float amplitude, float rings, float graininess, vec3 baseColor, vec3 ringColor)
{
    float woodPattern = WoodPattern(uv, frequency, amplitude, rings, graininess);
    vec3 color = mix(baseColor, ringColor, woodPattern);
    return color;
}



float ShadowCalculation(vec3 fragPos)
{
    vec3 fragToLight = fragPos - lightPositions[1];
    float closestDepth = texture(depthMap, fragToLight).r;
    closestDepth *= far_plane;
    
    float currentDepth = length(fragToLight);
   
    float shadow = 0.0;

    float bias = 0.0f; // max(0.02 * (1.0 - dot(vert.normal_world, fragToLight)), 0.005); 
    int samples = 20;
    float viewDistance = length(camera_world - fragPos);
    float diskRadius = (1.0 + (viewDistance / far_plane)) / 25.0;
    for(int i = 0; i < samples; ++i)
    {
        float closestDepth = texture(depthMap, fragToLight + gridSamplingDisk[i] * diskRadius).r;
        closestDepth *= far_plane;   // undo mapping [0;1]
        if(currentDepth - bias > closestDepth)
            shadow += 1.0;
    }
    shadow /= float(samples);
        
    return shadow;
    
}

vec3 fresnelSchlick(float cosTheta, vec3 F0)
{
    return F0 + (1.0 - F0) * pow(clamp(1.0 - cosTheta, 0.0, 1.0), 5.0);
}  


float DistributionGGX(vec3 N, vec3 H, float roughness)
{
    float a      = roughness*roughness;
    float a2     = a*a;
    float NdotH  = max(dot(N, H), 0.0);
    float NdotH2 = NdotH*NdotH;
	
    float num   = a2;
    float denom = (NdotH2 * (a2 - 1.0) + 1.0);
    denom = PI * denom * denom;
	
    return num / denom;
}

float GeometrySchlickGGX(float NdotV, float roughness)
{
    float r = (roughness + 1.0);
    float k = (r*r) / 8.0;

    float num   = NdotV;
    float denom = NdotV * (1.0 - k) + k;
	
    return num / denom;
}

float GeometrySmith(vec3 N, vec3 V, vec3 L, float roughness)
{
    float NdotV = max(dot(N, V), 0.0);
    float NdotL = max(dot(N, L), 0.0);
    float ggx2  = GeometrySchlickGGX(NdotV, roughness);
    float ggx1  = GeometrySchlickGGX(NdotL, roughness);
	
    return ggx1 * ggx2;
}



void main() {	

	//texture
	//vec3 albedo = texture(texture_albedo, vert.uv).rgb;
	//vec3 albedo = pow(vec3(texture(texture_albedo, vert.uv)), vec3(2.2));
	//float metallic = texture(texture_metallic, vert.uv).r;
	//float roughness = texture(texture_roughness, vert.uv).r;
	//float ao = texture(texture_ao, vert.uv).r;
    
    vec3 albedo = pow(vec3(texture(texture_albedo, vert.uv)), vec3(2.2));

    if (pattern == 1) {
        //albedo = WoodColorPattern(vert.uv, 2.0, 0.2, 10.0, 5.0, vec3(0.7), vec3(1.0)); } 
        //albedo = WoodColorPattern(vert.uv, 3.0, 0.3, 15.0, 5.0, vec3(0.6), vec3(1.0)); } 
        //albedo = WoodColorPattern(vert.uv, 0.10, 0.3, 55.0, 5.0, vec3(0.8), vec3(1.0)); } 
        albedo = WoodColorPattern(vert.uv, 3.0, 0.3, 40.0, 5.0, vec3(0.8), vec3(1.0)); } 
    else if (pattern == 2) {
        albedo = CheckerboardPattern(20.0f, vec3(1.0f), vec3(0.1f)); }
    else if (pattern == 3) {
        albedo = CoolPattern(20.0f, vec3(1.0f), vec3(0.1f)); }
    else if (pattern == 4) {
        albedo = OrganicPattern(vert.uv, texture_albedo); }
   

	//nontexture
	vec3 N = normalize(vert.normal_world);
	vec3 V = normalize(camera_world - vert.position_world);

	vec3 F0 = vec3(0.04);  //nonmetallic F0 is always 0.04
	F0 = mix(F0, albedo, metallic);

    //PBR
	vec3 Lo = vec3(0.0);
	for(int i = 1; i < 2; i++) // i < 4; ++i) 
	{
		vec3 L = normalize(lightPositions[i] - vert.position_world);
		vec3 H = normalize(V + L);
		float distance = length(lightPositions[i] - vert.position_world);
		float attenuation = 1.0 / (distance * distance);
		vec3 radiance = lightColors[i] * attenuation; 
    

		float NDF = DistributionGGX(N, H, roughness);
		float G = GeometrySmith(N, V, L, roughness);
		vec3 F = fresnelSchlick(max(dot(H, V), 0.0), F0);

		vec3 kS = F;
		vec3 kD = vec3(1.0) - kS;
		kD *= 1.0 - metallic;


		vec3 numerator = NDF * G * F;
		float denominator = 4.0 * max(dot(N, V), 0.0) * max(dot(N, L), 0.0)  + 0.0001;
		vec3 specular = numerator / denominator; 

		
		float NdotL = max(dot(N, L), 0.0);        
		Lo += (kD * albedo / PI + specular) * radiance * NdotL;
	
	}

    //shadows
    float shadow = ShadowCalculation(vert.position_world);
	vec3 ambient = vec3(0.3) * albedo * ao;
	vec3 color = ambient + Lo * (1 - shadow);  
	color = color / (color + vec3(1.0));
    color = pow(color, vec3(1.0/2.2));
	FragColor = vec4(color, 1.0) * brightness;
}

